using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;

namespace FinalDemo.Framework
{
	public class PropertyMonitor : IDisposable
	{
		private readonly ViewModelBase _propertySource;
		private readonly IDictionary<string, List<string>> _propertyMap;
		private readonly IDictionary<string, List<string>> _methodMap;
		private bool _disposed;

		public PropertyMonitor(ViewModelBase propertySource)
		{
			if (propertySource == null)
				throw new ArgumentNullException("propertySource");

			_propertyMap = MapDependencies<DependsOnAttribute>(() => propertySource.GetType().GetProperties());
			_methodMap = MapDependencies<DependsOnAttribute>(() => propertySource.GetType().GetMethods());
			VerifyDependancies(propertySource);

			_propertySource = propertySource;
			_propertySource.PropertyChanged += ProperySourcePropertyChanged;
		}

		private void ProperySourcePropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			if (_propertyMap.ContainsKey(e.PropertyName))
				_propertyMap[e.PropertyName].Each(_propertySource.NotifyPropertyChanged);

			if (_methodMap.ContainsKey(e.PropertyName))
				_methodMap[e.PropertyName].Each(ExecuteMethod);
		}

		private void ExecuteMethod(string name)
		{
			var memberInfo = _propertySource.GetType().GetMethod(name);
			if (memberInfo == null)
				return;

			memberInfo.Invoke(_propertySource, null);
		}

		public void Dispose()
		{
			if (_propertySource == null || _disposed)
				return;

			try
			{
				_propertySource.PropertyChanged -= ProperySourcePropertyChanged;
			}
			finally
			{
				_disposed = true;
			}
		}

		private static IDictionary<string, List<string>> MapDependencies<T>(Func<IEnumerable<MemberInfo>> getInfo) where T : DependsOnAttribute
		{
			var dependencyMap = getInfo().Distinct(new MemberInfoComparer()).ToDictionary(
							 p => p.Name,
							 p => p.GetCustomAttributes(typeof(T), true)
									 .Cast<T>()
									 .Select(a => a.DependencyName)
									 .ToList());

			return Invert(dependencyMap);
		}

		private static IDictionary<string, List<string>> Invert(IDictionary<string, List<string>> map)
		{
			var flattened = from key in map.Keys
								 from value in map[key]
								 select new { Key = key, Value = value };

			var uniqueValues = flattened.Select(x => x.Value).Distinct();

			return uniqueValues.ToDictionary(
							x => x,
							x => (from item in flattened
									where item.Value == x
									select item.Key).ToList());
		}

		private static void VerifyDependancies(object propertySource)
		{
			var methods = propertySource.GetType().GetMethods();
			var properties = propertySource.GetType().GetProperties();

			var propertyNames = methods.Union<MemberInfo>(properties)
				 .SelectMany(method => method.GetCustomAttributes(typeof(DependsOnAttribute), true).Cast<DependsOnAttribute>())
				 .Where(attribute => attribute.VerifyStaticExistence)
				 .Select(attribute => attribute.DependencyName);

			propertyNames.Each(x => VerifyDependancy(propertySource, x));
		}

		private static void VerifyDependancy(object propertySource, string propertyName)
		{
			var property = propertySource.GetType().GetProperty(propertyName);
			if (property == null)
				throw new ArgumentException("DependsOn Property Does Not Exist: " + propertyName);
		}

	}

	class MemberInfoComparer : IEqualityComparer<MemberInfo>
	{
		public bool Equals(MemberInfo x, MemberInfo y)
		{
			return x.Name == y.Name;
		}

		public int GetHashCode(MemberInfo obj)
		{
			return obj.Name.GetHashCode();
		}
	}

	public static class EventExtensions
	{
		public static void Each<T>(this IEnumerable<T> items, Action<T> action)
		{
			foreach (var item in items)
				action(item);
		}

		public static void Raise(this PropertyChangedEventHandler eventHandler, object source, string propertyName)
		{
			var handlers = eventHandler;
			if (handlers != null)
				handlers(source, new PropertyChangedEventArgs(propertyName));
		}
	}
}